#!/usr/bin/env python
# Copyright 2008-2009 David D Lowe
# Released under the GPLv3
"""Contains various functions for finding and loading Pigment objects from
installation directories"""

from __future__ import division
import const
from const import DEBUG, MY_CACHE_HOME, MY_DATA_HOME, PIGMENT_TYPES
from gettext import gettext as _
import sys, os
import handy
import threading
import gtk, gtk.glade
import gobject, subprocess
import urllib, socket, urlparse
import ep_exceptions
import shell

def find_skins(directories=None, pigments=None):
    """Looks for skins in the directories passed and returns a list of Skin objects
    and a list of exceptions encountered, similar to find_pigments
    directories: the list of directories to search for in order of priority, by default
    it is [MY_DATA_HOME, SHARED_PIGMENT_DIR]
    pigments: a list of pigments already loaded that Skin can use to search for pigments"""
    if directories is None:
        directories = [MY_DATA_HOME, const.SHARED_PIGMENT_DIR]
    retPigments = []
    pt = get_pigment_type("skin")().directoryName
    excps = []
    for dirr in directories:
        if os.path.exists(os.path.join(dirr, pt)):
            for pigmDir in os.listdir(os.path.join(dirr, pt)):
                if os.path.exists(os.path.join(dirr, pt, pigmDir, "pigment.xml")):
                    try:
                        retPigments.append(load_skin(pigmDir, [dirr], pigments=pigments))
                    except:
                        excps.append(list(sys.exc_info()) + [pigmDir])
    return retPigments, excps


def find_pigments(pigmentTypeStr, directories=None):
    """Looks for pigments in the directories passed and returns a list of
    Pigment objects, and a list of tuples of exception information
    that may have been encountered loading pigments, hopefully, just [] ;)
    This format: [pigment1, pigment2], [FileNotFoundException, exception, traceback, pigmDir]
    pigmentTypeStr: the pigment type to look for, string (example: "wallpaper")
    directories: a list of directories to search for in order of priority, by default it is
        [MY_DATA_HOME, SHARED_PIGMENT_DIR]"""
    if directories is None:
        directories = [MY_DATA_HOME, const.SHARED_PIGMENT_DIR]
    retPigments = []
    excps = []
    pt = get_pigment_type(pigmentTypeStr)()
    for dirr in directories:
        if os.path.exists(os.path.join(dirr, pt.directoryName)):
            for pigmDir in os.listdir(os.path.join(dirr, pt.directoryName)):
                if os.path.exists(os.path.join(dirr, pt.directoryName, pigmDir, "pigment.xml")):
                    try:
                        retPigments.append(load_pigment(pigmDir, pigmentTypeStr, [dirr]))
                    except Exception, ee:
                        excps.append(list(sys.exc_info()) + [os.path.join(dirr, pt.directoryName,pigmDir)])
    return retPigments, excps

def load_pigment(pigment, pigmentType, directories=None):
    """Returns a Pigment object that has been read from the appropriate file
    pigment: the pigment's codeName
    pigmentType: its type, string
    directories: list of installationDirectories to search through, in
    order of priority. By default it is [MY_DATA_HOME, SHARED_PIGMENT_DIR]"""
    if directories is None:
        directories = [MY_DATA_HOME, const.SHARED_PIGMENT_DIR]
    pp = create_pigment(pigmentType)
    if isinstance(directories, str):
        raise(Exception("deprecated"))
        directories = [directories]
    for dir in directories:
        if does_pigment_exist(pigment, pigmentType, [dir]):
            pp.installationDirectory = dir
            pp.read(os.path.join(dir, get_pigment_type(pigmentType).directoryName, pigment, "pigment.xml"))
            pp.installationDirectory = dir
            return pp
    raise(ep_exceptions.PigmentNotFoundException(pigment, pigmentType, directories))

def load_skin(skin, directories=None, pigments=None):
    """Returns a Skin object that has been read from the appropriate file
    skin: the skin's codename
    directories: list of installationDirectories to search through, in order of
    priority. By default it is [MY_DATA_HOME, SHARED_PIGMENT_DIR]
    pigments: list of preloaded Pigment objects"""
    if directories is None:
        directories = [MY_DATA_HOME, const.SHARED_PIGMENT_DIR]
    if pigments is None: pigments = []
    pp = create_pigment("skin")
    if isinstance(directories, str):
        raise(Exception("deprecated"))
        directories = [directories]
    for dir in directories:
        if does_pigment_exist(skin, "skin", [dir]):
            pp.installationDirectory = dir
            pp.read_with_pigments(os.path.join(dir, get_pigment_type("skin")().directoryName, skin, "pigment.xml"), 
                pigments=pigments)
            pp.installationDirectory = dir
            return pp
    raise(ep_exceptions.PigmentNotFoundException(skin, "skin", directories))

def does_pigment_exist(pigmentStr, pigmentType, directories):
    """Returns whether the pigment specified is installed in the installationDirectory
    pigmentStr: a string specifying the pigment's code name
    pigmentType: a string specifying the codeName of the pigment's type
    directories: a list of installationDirectories in order of priority in which to search through"""
    if directories is None:
        directories = [MY_DATA_HOME, const.SHARED_PIGMENT_DIR]
    if isinstance(directories, str):
        raise(Exception("deprecated"))
    for dir in directories:
        if os.path.exists(os.path.join(dir, get_pigment_type(pigmentType)().directoryName, pigmentStr, "pigment.xml")):
            return True
        else:
            return False

def get_pigment_type(pigmentTypeString):
    """Return the appropriate PigmentType class based on the string argument
    Note: this includes the PackType, even though Pack is not a subclass of Pigment"""
    
    from pigments import skin, wallpaper, metacity_, gtk_, icons, gnomesplash, cursors, grub, gdm, xsplash_hack, pack
       
    pts = {"skin":skin.SkinType, "wallpaper":wallpaper.WallpaperType, "metacity": metacity_.MetacityType, \
        "gtk":gtk_.GTKType, "icons":icons.IconsType, "gnomeSplash":gnomesplash.GnomeSplashType, \
        "cursors":cursors.CursorsType, "grub":grub.GrubType, "gdm":gdm.GDMType, \
        "xsplashHack":xsplash_hack.XsplashHackType, "pack":pack.PackType}
        
    if pigmentTypeString in pts.keys():
        return pts[pigmentTypeString]
    else:
        raise(ep_exceptions.PigmentTypeNotFoundException(pigmentTypeString))

def create_pigment(pigmentType):
    """Creates a new Pigment of type pigmentType
    For example, passing "wallpaper" to this function would return wallpaper.Wallpaper()"""
    
    if pigmentType in const.PIGMENT_TYPES:
        return get_pigment_type(pigmentType)().get_Pigment()()
    else:
        raise(ep_exceptions.PigmentTypeNotFoundException(pigmentType))

def manage_repo_pigments(application, installDict, removeDict, installationDirectory=None):
    """manage_repo_piments installs and uninstalls pigments! It also handles the GUI side of things
    
    application: the EpidermisApp object that's being run, usually self
    installDict: a dictionary of the pigment codeNames to install from the repo, example:
        {"skin":[],"wallpaper":["wallpigment1"],"metacity":[],"gtk":["gtkpak1","gtkpak2"], \
        "icons":[],"gnomeSplash":[],"cursors":[],"grub":[],"gdm":[],"xsplashHack":[]}
    removeDict: a dictionary of the pigments codeNames to uninstall, in the same format as installDict
    installationDirectory: the base directory where to install pigments
    Should be called in the main thread"""
    
    
    application.runningManagePigments += 1
    if application.runningManagePigments > 1:
        raise()
    
    
    def cancel_callback(widget=None):
        """cancel_callback for the cancel button on the download window"""
        app.cancel = True
        gobject.idle_add(lambda:fetchWin.hide())
        
    def fetchWin_delete_callback(widget, event):
        """the delete_callback for the download window, calls the cancel_callback, 
        disallows closing"""
        cancel_callback(cancelButton)
        return True
    
    # still under manage_repo_piments
    
    app = application
    # this next line reloads the glade file afresh, creating new download window
    wTree = gtk.glade.XML(app.wTreeFileName, root="winOperations")
    
    # set up window
    fetchWin = wTree.get_widget("winOperations")
    fetchWin.set_destroy_with_parent(True)
    fetchWin.set_title("Installing and uninstalling pigments")
    # TODO: set parent
    progressbar = wTree.get_widget("progressbarOperations")
    cancelButton = wTree.get_widget("butOperationsCancel")
    progressbar.set_fraction(0)
    fetchWin.connect("delete-event", fetchWin_delete_callback)
    app.cancel = False
    cancelButton.connect("clicked", cancel_callback)
    
    # start thread and show window
    thr = ManageRepoPigmentsThread(installDict,removeDict,app,progressbar,wTree, fetchWin, installationDirectory)
    fetchWin.show_all()
    thr.start()

class ManageRepoPigmentsThread(threading.Thread):
    """ManageRepoPigmentsThread installs and uninstalls pigments"""
    
    def __init__(self,installDict,removeDict,app,progressbar, wTree, fetchWin, installationDirectory=None):
        """installDict: a dictionary of the pigments codeNames to install, example:
            {"skin":[],"wallpaper":["wallpigment1"],"metacity":[],"gtk":["gtkpak1","gtkpak2"], \
            "icons":[],"gnomeSplash":[],"cursors":[],"grub":[],"gdm":[],"xsplashHack":[]}
        removeDict: a dictionary of the pigments codeNames to uninstall, similar to installDict in format
        app: the EpidermisApp application object
        progressbar: the progressbar where progress should be displayed
        wTree: the wTree from which fetchWin was referenced
        fetchWin: the window where progress should be displayed
        installationDirectory: base directory where to install pigments"""
        self.installDict = installDict
        self.removeDict = removeDict
        self.app = app
        self.progressbar = progressbar
        self.wTree = wTree
        self.fetchWin = fetchWin
        self.installationDirectory=installationDirectory
        if self.installationDirectory is None: self.installationDirectory = const.SHARED_PIGMENT_DIR
        self.urls = []
        self.operationsSummary = self.wTree.get_widget("lblOperationsSummary")
        self.operationsInformation = self.wTree.get_widget("lblOperationsInformation")
        threading.Thread.__init__(self)
        self.setName("ManageRepoPigmentsThread")
        
    def run(self):
        
        gobject.idle_add(lambda:self.operationsSummary.set_text(_("Installing and uninstalling pigments")))
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("You may need to enter a password to gain root access")))

        # initialize a BashRootShell
        shll = shell.BashRootShell()
        
        # count number of pigments to install, if more than one, call
        # download_pigments and install_pigments
        npti = 0
        for ii in self.installDict:
            npti += len(self.installDict[ii])
        if npti > 0:
            shll.prepare()
            self.download_pigments()
            self.install_pigments(shll)
        
        
        #count number of pigments to uninstall, if more than one, call remove_pigments
        nptu = 0
        for ii in self.removeDict:
            nptu += len(self.removeDict[ii])
        if nptu > 0:
            if not shll.ready:
                shll.prepare()
            self.remove_pigments(shll)
        
        # hide window
        gobject.idle_add(lambda: self.fetchWin.hide())
        
        # exit shell
        if shll.ready:
            shll.exit()
        
        # find and store customised skin
        customisedSkin = self.app.customisedSkin
        
        # refresh app's main view:
        gobject.idle_add(lambda:self.app.find_and_load_installed_pigments())
        gobject.idle_add(lambda:self.app.add_skins())
        
        # add customised skin again
        # make sure any linked pigments have not been uninstalled
        removeCustomisedSkin = False
        if not customisedSkin is None:
            for pt in customisedSkin.linkedPigments:
                if not does_pigment_exist(customisedSkin.linkedPigments[pt].codeName, pt):
                    removeCustomisedSkin = True
                    break
            if not removeCustomisedSkin:
                gobject.idle_add(lambda:self.app.append_skin_to_listStore(customisedSkin, True))
        
        # refresh repo
        if (npti > 0) | (nptu > 0):
            import loadrepo
            gobject.idle_add(lambda:loadrepo.load_repo(self.app, forceReload = True))
        
        self.app.runningManagePigments -= 1
        gobject.idle_add(lambda:self.app.controls.btInstallPigments.set_sensitive(True))
        
        
    def download_pigments(self):
        """downloads the pigments in self.installDict and moves them to
        MY_CACHE_HOME/pigments/pigmentType/pigmentName_type.pigment"""
        
        gobject.idle_add(lambda:self.operationsSummary.set_text(_("Downloading pigments to install")))
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("The pigments you selected are being downloaded from the repository")))
        
        # prepare list of urls and prepare directories in myCacheHome
        self.urls = []
        repoDir = self.app.repoURL[:(-1*len(self.app.repoURL.split("/")[-1]))]
        for pt in self.installDict:
            for pk in self.installDict[pt]:
                self.urls.append(repoDir + get_pigment_type(pt).directoryName + "/" + pk + "/" + pk + "_" + pt + ".pigment")
        subprocess.call(["mkdir", "-p", os.path.join(MY_CACHE_HOME,"pigments",get_pigment_type(pt).directoryName)])
        
        # download the urls one by one
        for url in self.urls:
            gtk.gdk.threads_enter()
            self.progressbar.set_text(_("Downloading %s") % url.split("/")[-1:][0])
            downloadingText = _("Downloading %s") % url.split("/")[-1:][0]
            self.progressbar.set_fraction(0)
            gtk.gdk.threads_leave()
            def download_hook(blocksDownloaded, blockSize, size):
                """download_hook for the download thread that displays progress
                in the progressbar"""
                downloaded = blocksDownloaded * blockSize
                frac = blocksDownloaded * blockSize/ size
                if frac > 1:
                    frac = 1
                    downloaded = size
                if frac < 0: frac = 0
                gtk.gdk.threads_enter()
                self.progressbar.set_fraction(frac)
                if size >= 0:
                    self.progressbar.set_text(downloadingText + " " + handy.bytes_in_str(downloaded) + "/" + handy.bytes_in_str(size))
                else:
                    self.progressbar.set_text(downloadingText + " " + handy.bytes_in_str(downloaded) + "/??")
                gtk.gdk.threads_leave()
            filename, headers = urllib.urlretrieve(url, reporthook=download_hook)
            # check to see if server returned a 404 html message
            downloadedFile = open(filename, 'r')
            head = downloadedFile.read(256)
            if "<html" in head and "404 - Not Found" in head:
                downloadedFile.close()
                raise(ep_exceptions.UrlNotFoundException(url))
            downloadedFile.close()
            # move downloaded file to MY_CACHE_HOME/pigments/pigmentType/pigmentName_type.pigment
            subprocess.call(["mkdir", "-p", os.path.join(MY_CACHE_HOME, "pigments", url.split("/")[-3:-2][0], url.split("/")[-1].split(".")[0].split("_")[0])])
            subprocess.call(["mv", "-f",filename, \
            os.path.join(MY_CACHE_HOME,"pigments",url.split("/")[-3:-2][0],url.split("/")[-1].split(".")[0].split("_")[0], url.split("/")[-1])])
            if not os.path.exists(os.path.join(MY_CACHE_HOME,"pigments",url.split("/")[-3:-2][0],url.split("/")[-1].split(".")[0].split("_")[0], url.split("/")[-1])):
                raise(ep_exceptions.FileNotFoundException(os.path.join(MY_CACHE_HOME,"pigments",url.split("/")[-3:-2][0],url.split("/")[-1].split(".")[0].split("_")[0], url.split("/")[-1])))
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("Downloading finished")))

        
    def install_pigments(self, bashShell):
        """installs pigments in self.installDict by running managepigments.py 
        as root with gksu
        bashSholl: a prepared BashRootShell"""
        
        gobject.idle_add(lambda:self.operationsSummary.set_text(_("Installing pigments")))
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("The pigments you selected are being installed")))
        
        shll = bashShell
        if not shll.ready:
            raise(shell.ShellNotReadyException())
        
        def progressCallback(progress, query=None):
            gtk.gdk.threads_enter()
            self.progressbar.set_fraction(progress)
            gtk.gdk.threads_leave()
            return self.app.cancel        
        
        for pt in PIGMENT_TYPES[1:] + ["skin"]:
            for pigmCodeName in self.installDict[pt]:
                gtk.gdk.threads_enter()
                self.progressbar.set_text(_("Installing %s") % pigmCodeName + "_" + pt)
                self.progressbar.set_fraction(0)
                gtk.gdk.threads_leave()

                pigm = create_pigment(pt)
                pigm.load(os.path.join(MY_CACHE_HOME, "pigments", get_pigment_type(pt).directoryName, \
                    pigmCodeName, pigmCodeName + "_" + pt + ".pigment"))
                pigm.install(self.installationDirectory, shell=shll, progressCallback=progressCallback)
                
        # finishing touches
        shll.do(["chown", "-R", "%s:%s" % (const.USERNAME, const.USERNAME), MY_DATA_HOME])
        
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("Installing pigments done")))

    
    def remove_pigments(self, bashShell):
        """uninstalls pigments found in self.removeDict by running managepigments.py
        as root with gksu"""
        

        gobject.idle_add(lambda:self.operationsSummary.set_text(_("Uninstalling pigments")))
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("The pigments you selected are being removed")))

        def progressCallback(progress, query=None):
            gtk.gdk.threads_enter()
            self.progressbar.set_fraction(progress)
            gtk.gdk.threads_leave()
            return self.app.cancel     
        
        for pt in PIGMENT_TYPES: # skins first!
            for pigmCodename in self.removeDict[pt]:
                # search for the installed pigment object
                for installedPigm in self.app.installedPigments[pt]:
                    found = False
                    if installedPigm.codeName == pigmCodename:
                        gobject.idle_add(lambda:self.operationsInformation.set_text(_("Uninstalling %s") % pigmCodename + "_" + pt))
                        installedPigm.remove(progressCallback=progressCallback, shell=bashShell)
                        found = True
                        break
                if not found:
                    print >> sys.stderr, "Could not uninstall pigment %s of type %s" % (pigmCodename, pt)
    
        
        gobject.idle_add(lambda:self.operationsInformation.set_text(_("Uninstallation finished")))

def main():
    print "finding pigments"
    for pt in PIGMENT_TYPES[1:]:
        for pigm in find_pigments(pt)[0]:
            print pigm

if __name__ == "__main__":
    main()
